=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright 2012 Fredo6 - Designed and written June 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6Material.rb
# Original Date	:   27 Feb 2012
# Description	:   Manage Materials
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#=============================================================================================
#=============================================================================================
# Class MaterialManager: main class for the management of Materials
#=============================================================================================
#=============================================================================================

class MaterialManager

#-------------------------------------------------------------------
# Initialization
#-------------------------------------------------------------------

#Initialization
def initialize__(*hargs)
	@hsh_mat = {}
	@tw = Sketchup.create_texture_writer
	@texture_dir = File.join LibFredo6.tmpdir, "LibFredo6_Textures"
	Dir.mkdir @texture_dir unless FileTest.directory?(@texture_dir)
	
	#Parsing the arguments
	hargs.each { |arg| arg.each { |key, value| parse_args(key, value) } if arg.class == Hash }

	#Creating the observers
	Sketchup.add_observer self
	model_changing Sketchup.active_model, 'Boot'
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /notify_proc/i
		@notify_proc = value
	end	
end	

#-------------------------------------------------------------------
# Observer methods
#-------------------------------------------------------------------

#Observers events
def onNewModel(model) ; model_changing(model, 'new') ; end
def onOpenModel(model) ; model_changing(model, 'open') ; end

#New model or Opening the model
def model_changing(model, mode)
	model.materials.add_observer self
	history_initialize
end

def onMaterialSetCurrent(materials, curmat)
	notify_change
end

def onMaterialChange(materials, curmat)

end

def onMaterialRefChange(materials, curmat)

end

#Notification of a change of material by the user
def notify_change
	#Do nothing as change is initiated by MaterialManager
	if @setting_current
		@setting_current = false
		return
	end
	
	#Substitute a material already in the model if applicable
	curmat = current()
	mat = is_material_loaded?(curmat)
	if mat != false && mat != curmat
		UI.start_timer(0) { load_material mat }
	end	
	
	#Notify the caller
	notify_caller :change_current
	
	#Store in the stack of recent materials
	history_store mat if mat && mat.valid?
end

def notify_caller(event)
	return unless @notify_proc
	@notify_proc.call event
end

#-------------------------------------------------------------------
# Managing Current Material
#-------------------------------------------------------------------

#Get current material
def current
	Sketchup.active_model.materials.current
end

def current=(mat)
	Sketchup.active_model.materials.current = mat
end

def private_set_current(mat)
	@setting_current = true
	Sketchup.active_model.materials.current = mat
end

def current_info
	mat = Sketchup.active_model.materials.current
	if mat
		info = mat.display_name
		tx = mat.texture
		if tx
			w = Sketchup.format_length tx.width
			h = Sketchup.format_length tx.height
			info += " - w = #{w} h = #{h}"
		end	
	else
		info = T6[:T_TXT_DefaultMaterial]
	end
	info
end

def current_name
	material_name Sketchup.active_model.materials.current
end

def material_name(mat)
	name = (mat && mat.valid?) ? mat.name : T6[:T_TXT_DefaultMaterial]
	return "" unless name
	name = name.gsub('[', '')
	name = name.gsub(']', '')
	name
end

#Check if current material has texture
def current_with_texture?
	curmat = Sketchup.active_model.materials.current
	(curmat && curmat.texture) ? true : false
end

#---------------------------------------------------------------------------
# Loading of material
#---------------------------------------------------------------------------

#Check if a material is loaded
#Default material (curmat == nil) is always loaded
#return the material which is already loaded or false
#test must be done with false: is_loaded?(mat) == false
def is_material_loaded?(curmat)
	return curmat unless curmat
	materials = Sketchup.active_model.materials
	
	#Part of the list of materials
	return curmat if curmat && materials.to_a.include?(curmat)
	
	#Checking if a material has the same attributes in the materials of the model
	materials.each do |mat|
		next if mat.materialType != curmat.materialType || mat.alpha != curmat.alpha 
		next unless color_same?(mat.color, curmat.color)
		curtx = curmat.texture
		tx = mat.texture
		next if (curtx && !tx) || (!curtx && tx)
		if tx && curtx
			next if File.basename(tx.filename) != File.basename(curtx.filename)
			next if tx.image_width != curtx.image_width || tx.image_height != curtx.image_height
			next if tx.width != curtx.width || tx.height != curtx.height
		end	
		return mat
	end
	false
end
	
def material_print(mat, text)
	puts "\nText = #{text}"
	puts "Name = #{mat.name}"
	puts "Display Name = #{mat.display_name}"
	puts "Color = #{mat.color.to_a.inspect} i = #{mat.color.to_i}"
	puts "Alpha = #{mat.alpha}"
	puts "Type = #{mat.materialType}"
	tx = mat.texture
	if tx
		puts "Filename = #{File.basename tx.filename}"
		puts "image = #{tx.image_width} - #{tx.image_height}"
		puts "image = #{tx.width} - #{tx.height}"
	end	
end
	
def color_same?(color1, color2)
	lc1 = color1.to_a
	lc2 = color2.to_a
	for i in 0..2
		return false if (lc1[i] - lc2[i]).abs > 2
	end
	true
end
	
#Load and set the current material into the model
def load_material(curmat=false, ops_text=nil)
	model = Sketchup.active_model
	materials = model.materials
	curmat = materials.current if curmat == false
	newmat = is_material_loaded?(curmat)
	
	#Already loaded
	unless newmat == false
		private_set_current newmat
		return newmat
	end	

	#Cloning the material without a texture
	tx = curmat.texture
	unless tx
		newmat = materials.add curmat.name
		newmat.alpha = curmat.alpha
		newmat.color = curmat.color
		private_set_current newmat
		return newmat
	end	
	
	#Loading the material with a texture - Use a fake group and abort
	t0 = Time.now.to_f
	
	#Getting the Texture file of the current material
	tx_fullpath = tx.filename
	
	if FileTest.exist?(tx_fullpath)
		new_txpath = tx_fullpath
		mpath = nil
		
	#Applying it to a fake group - This allows to copy the texture imgae to the temp directory
	else	
		texture_path = File.basename tx_fullpath
		mpath = File.join @texture_dir, texture_path
		ops_text = "LibFredo6 Texture" unless ops_text
		model.start_operation ops_text
		g = model.active_entities.add_group 
		g.material = curmat
		@tw.load g
		@tw.write g, mpath
		model.abort_operation
		new_txpath = mpath
	end	
	
	#Adding the material
	newmat = materials.add curmat.name
	newmat.alpha = curmat.alpha
	newmat.color = curmat.color
	newmat.texture = new_txpath
	newtx = newmat.texture
	newtx.size = [tx.width, tx.height]
	File.delete mpath if mpath
	
	private_set_current newmat
	newmat
end

#-------------------------------------------------------------------
# Material History Management
#-------------------------------------------------------------------

#Clean up after usage
def history_initialize
	@lst_history = []
	@ipos_history = 0
end

def history_store(mat)
	@lst_history = @lst_history.find_all { |m| m && m.valid? }
	return unless mat && mat.valid?
	@lst_history.delete mat
	@lst_history.push mat
end

def history_remove(mat)
	@lst_history.delete mat
end

def history_navigate(incr, beg_end=false)
	lsm = history_build_list
	curmat = Sketchup.active_model.materials.current
	newmat = is_material_loaded?(curmat)
	curmat = newmat if newmat != false && newmat != curmat
	n = lsm.length
	if beg_end
		#ipos = (incr > 0) ? n-1 : n - @lst_history.length
		ipos = (incr > 0) ? n-1 : n - @lsm.length
	else
		if RUN_ON_MAC
			ipos = @ipos_history
		else
			ipos = lsm.index(curmat)
			ipos = n-1 unless ipos
		end
		ipos += incr
	end	
	ipos = ipos.modulo(n)
	@ipos_history = ipos
	load_material lsm[ipos]
end

def history_build_list
	lsm = []
	materials = Sketchup.active_model.materials
	for i in 0..materials.count-1
		lsm.push materials[i]
	end
	lsm = lsm.sort { |a, b| a.display_name <=> b.display_name }
	#@lst_history.each { |mat| lsm.delete mat }
	lsm + @lst_history
end

end	#class MaterialManager

#=============================================================================================
#=============================================================================================
# Class MaterialPicker: main class for the management of Material Picking in the model
#=============================================================================================
#=============================================================================================

T6[:TIT_MaterialPicker_TopMessage] = "SAMPLE A MATERIAL IN MODEL"	
T6[:TIT_MaterialPicker_StatusMessage] = "Pick a face to sample its material"

class MaterialPicker

#-------------------------------------------------------------------
# Initialization
#-------------------------------------------------------------------

#INIT: Class instance Initialization
def initialize__(*hargs)
	@model = Sketchup.active_model
	@view = @model.active_view
	@ph = @view.pick_helper	
	@tr_id = Geom::Transformation.new
	@picked_material = nil
	@name_material = nil
	
	#Parsing the arguments
	hargs.each { |arg| arg.each { |key, value| parse_args(key, value) } if arg.class == Hash }
	
	#Initialize texts and cursors
	@id_cursor_pick = Traductor.create_cursor "Cursor_Sample_Material", 2, 2
	init_messages
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /notify_proc/i
		@notify_caller_proc = value
	end	
end	

#INIT: Text instruction for picking a material
def init_messages
	@ogl = Traductor::OpenGL_6.new
	bk_color = 'palegoldenrod'
	fr_color = 'yellow'
	sizey = 16
	sizex = 28
	fac = 2
	pts = G6.pts_rectangle 0, -0.5, sizex, sizey
	@pick_instructions_icon = [[GL_POLYGON, pts, bk_color], [GL_LINE_LOOP, pts, fr_color, 2, '']]
	@pick_instructions_icon += @ogl.draw_proc(:std_sample_material, sizex, sizey)
	text = T6[:TIT_MaterialPicker_TopMessage]
	x = fac * sizex + 4
	y = 2
	@pick_instructions = G6.rectangle_text_instructions(text, x, y, bk_color, fr_color, 0, 2, 0)
	ts = Geom::Transformation.scaling ORIGIN, fac, -fac, 1
	tt = Geom::Transformation.translation Geom::Vector3d.new(0, fac * sizey, 0)
	@tr_instructions = tt * ts
	@status_message = T6[:TIT_MaterialPicker_StatusMessage]
	@tip_default_material = T6[:T_TXT_DefaultMaterial]
end

#---------------------------------------------------------------------------------------------
# MODE: Manage the picking mode
#---------------------------------------------------------------------------------------------

def picking_start
	@picking_mode = true
	notify_caller :enter_picking
end

def picking_stop
	@picking_mode = false
	notify_caller :exit_picking
end

def picking?
	@picking_mode
end

def picked_material
	@picked_material
end
	
#MODE: Call the notification method of the caller class	
def notify_caller(event, default=nil)
	return default unless @notify_caller_proc
	val = @notify_caller_proc.call(event)
	return default if val == :no_event
	val
end

def show_message
	Sketchup.set_status_text "", SB_VCB_LABEL
	Sketchup.set_status_text "", SB_VCB_VALUE
	Sketchup.set_status_text @status_message
end
	
#---------------------------------------------------------------------------------------------
# MOVE: Mouse Movement Methods
#---------------------------------------------------------------------------------------------

#VIEWPORT: Computing Current cursor
def onSetCursor
	UI.set_cursor @id_cursor_pick
end

#MOVE: Mouse Movements
def onMouseMove(flags, x, y, view)
	@x = x
	@y = y
	@picked_material = nil
	@name_material = nil
	
	#Check the face under the mouse
	face_under_mouse(flags, x, y, view)
	
	#Identifying the material
	identify_material(x, y)
end

#MOVE: Compute the face under the mouse
def face_under_mouse(flags, x, y, view)
	#Picking the point
	@ph.do_pick x, y, 8	
	@picked_face = @ph.picked_face
	@tr = @tr_id	
	return nil unless @picked_face
	
	#Getting the transformation for the face
	for i in 0..@ph.count
		ls = @ph.path_at(i)
		if ls && ls.include?(@picked_face)
			@tr = @ph.transformation_at(i)
			break
		end
	end	
end

#MOVE: Identify the material on the face
def identify_material(x, y)
	return unless @picked_face
	
	#Check if face is normal or reversed
	#ray = @view.pickray(x, y)
	#vec = @tr * @picked_face.normal
	@picked_material = (@tr * @picked_face.normal % @view.pickray(x, y)[1] <= 0) ? @picked_face.material : @picked_face.back_material
	
	#Name of the material
	@name_material = (@picked_material) ? @picked_material.display_name : @tip_default_material	
end

#---------------------------------------------------------------------------------------------
# CLICK: Click Management
#---------------------------------------------------------------------------------------------

#THRUPAINT_TOOL - MATERIAL: Play a camera click sound
def play_camera_sound	
	@camera_sound = Traductor::MYPLUGIN.picture_get "Camera_click.wav" unless @camera_sound
	UI.play_sound @camera_sound if FileTest.exist?(@camera_sound)
end

#Sample the material and exit
def sample_material
	return unless @picked_face
	picking_stop if @picked_face
	play_camera_sound
	notify_caller :material
end

#Return key pressed
def onReturn(view)
	sample_material
end

#CLICK: Button click DOWN
def onLButtonDown(flags, x, y, view)
	@button_down = true
	sample_material
end

#CLICK: Button click UP - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	return unless @button_down
	@button_down = false
end

#CLICK: Double Click received
def onLButtonDoubleClick(flags, x, y, view)
	picking_stop	
end

#---------------------------------------------------------------------------------------------
# DRAW: Drawing Methods
#---------------------------------------------------------------------------------------------

#DRAW: Draw top method
def draw(view)	
	@ogl.process_draw_GL view, @tr_id, @pick_instructions
	@ogl.process_draw_GL view, @tr_instructions, @pick_instructions_icon
	G6.draw_rectangle_text(view, @x, @y, @name_material, 'lightgreen', 'green') if @x && @name_material
end

end	#class MaterialPicker

end	#End Module Traductor
