=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# © Copyright May 2014
# Designed and developped May 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Lib6Stencil.rb
# Original Date	:   18 May 2014
# Description	:   Stencil description for intersections
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#=============================================================================================
#=============================================================================================
# Class Stencil: 2D shape implementation
#=============================================================================================
#=============================================================================================

class Stencil

#---------------------------------------------------------------------------------------------
# PseudoShape: Pseudo face structure
#---------------------------------------------------------------------------------------------

PseudoShape = Struct.new :main_loop, :loops, :polys, :interiors, :centroid, :anchor

#---------------------------------------------------------------------------------------------
# INIT: Initializations
#---------------------------------------------------------------------------------------------

#INIT: Initialize the Stencil
def initialize__(*hargs)
	@model = Sketchup.active_model
	@entities = @model.active_entities
	@lst_pshapes = []
end

#INIT: Initialize the Stencil
def configure(from, within_operation, *hargs)
	@from = from
	@within_operation = within_operation
	@hsh_from = {}
	hargs.each { |harg| @hsh_from.update harg if harg.class == Hash }
	configure_from from
end

#INIT: Configure a custom shape
def configure_from(from)
	case from
	
	#From specification of loops
	when :from_specs
		@pseudoshapes_specs = @hsh_from[:from_specs]
		return unless @pseudoshapes_specs
		@pseudoshapes_specs.each do |a|
			pseudoshape_create *a
		end	
		
	#From faces in model
	when :from_faces
		@model.start_operation "Stencil creation from faces" unless @within_operation
		lst_faces = @hsh_from[:lst_faces]
		trf = @hsh_from[:tr_faces]
		repere_tr = @hsh_from[:repere_tr]
		build_from_faces(lst_faces, trf, repere_tr)
		@model.abort_operation unless @within_operation

	#From Grouponents in model
	when :from_comp
		@model.start_operation "Stencil creation from Group / Component" unless @within_operation
		comp = @hsh_from[:comp]
		trf = @hsh_from[:tr_comp]
		repere_tr = @hsh_from[:repere_tr]
		entities = G6.grouponent_entities(comp)
		lst_faces = entities.grep(Sketchup::Face)
		build_from_faces(lst_faces, trf, repere_tr)
		@model.abort_operation unless @within_operation
	
	end	
	
	#Compute the bounding box
	bbox_compute
end

#INIT: Check if the Stencil contains shapes
def valid?
	@lst_pshapes.length > 0
end

#---------------------------------------------------------------------------------------------
# BUILD: Build stencil from faces and grouponents
#---------------------------------------------------------------------------------------------

def build_from_faces(lst_faces, trf, repere_tr)
	t = repere_tr.inverse * trf
		
	g = @entities.add_group
	gent = g.entities
	
	#Sorting the faces based on number of loops
	lst_faces_simple = []
	lst_faces_hole = []
	lst_faces.each do |face|
		if face.loops.length == 1
			lst_faces_simple.push face
		else
			lst_faces_hole.push face
		end
	end
	lst_faces = lst_faces_simple + lst_faces_hole
	
	#Creating projected faces
	lst_holes = []
	pts_interiors = []
	plane = [ORIGIN, Z_AXIS]
	lst_faces.each do |face|
		face.loops.each_with_index do |loop, iloop|
			pts = [] 
			loop.vertices.collect do |vx| 
				pt = (t * vx.position).project_to_plane(plane)
				pts.push pt
				curve = vx.curve_interior?
				pts_interiors.push pt if curve && !curve.is_polygon?
			end	
			next if G6.curl_is_aligned?(pts)
			begin
				f = gent.add_face pts
			rescue
				next
			end
			lst_holes.push pts if iloop > 0
		end
	end
	
	#Cleaning up the coplanar edges
	lerase = gent.grep(Sketchup::Edge).find_all { |edge| G6.edge_coplanar?(edge) }
	gent.erase_entities lerase if lerase.length > 0
	
	#Creating the face holes
	lerase = []
	lst_holes.each do |pts|
		f = gent.add_face pts
		lerase.push f
	end
	lerase = lerase.find_all { |f| f.valid? }
	gent.erase_entities lerase if lerase.length > 0
	
	#Checking the remaining faces to build the pseudo shapes specifications
	pseudoshapes_specs = []
	lfaces = gent.grep(Sketchup::Face)
	lfaces.each do |face|
		loop_pts = []
		interiors = []
		face.loops.each_with_index do |loop, iloop|
			vertices = (iloop == 0) ? loop.vertices : loop.vertices.reverse
			pts = []
			interior_flags = []
			vertices.each do |vx|
				next if G6.lonely_vertex?(vx)
				ptvx = vx.position
				pts.push ptvx
				int = pts_interiors.find { |pt| pt == ptvx }
				interior_flags.push((int) ? true : false)
			end
			loop_pts.push pts
			interiors.push interior_flags
		end
		
		#Decomposition in polygons for painting
		polys = []
		mesh = face.mesh
		pts = mesh.points
		mesh.polygons.each { |p| polys.push p.collect { |i| pts[i.abs-1] } }
		pseudoshapes_specs.push [loop_pts, polys, interiors]
	end	
	
	#Erasing the temporary group
	g.erase!
	
	#Updating the stencil
	pseudoshapes_specs.each do |a|
		pseudoshape_create *a
	end		
end
	
#---------------------------------------------------------------------------------------------
# PSEUDOSHAPE: Manage the individual shape constituting the stencil
#---------------------------------------------------------------------------------------------

#PSEUDOSHAPE: Create a pseudo face structure
def pseudoshape_create(loops, polys=nil, interiors=nil)
	pshape = PseudoShape.new
	pshape.loops = loops
	pshape.main_loop = loops.first
	polys = [pshape.main_loop] unless polys
	pshape.polys = polys
	unless interiors
		interiors = []
		for i in 0..loops.length-1
			interiors[i] = []
		end
	end	
	pshape.interiors = interiors
	pshape.centroid, = G6.loop_centroid_area(pshape.main_loop, Z_AXIS)
	@lst_pshapes.push pshape
	pshape
end

#---------------------------------------------------------------------------------------------
# INFO: Information on Stencil
#---------------------------------------------------------------------------------------------

#INFO: Return the list of pseudoshapes
def pseudoshape_list
	@lst_pshapes
end

#INFO: Return the loop information
def pseudoshape_get_anchor(pshape)
	anchor = pshape.anchor
	(anchor) ? anchor : pshape.centroid
end

#INFO: Return the loop information
def pseudoshape_get_loops_info(pshape)
	loops_info = [] 
	interiors = pshape.interiors
	loops_info.push [pshape.main_loop.clone, :main, interiors[0]]
	#pshape.loops[1..-1].each_with_index { |loop, i| loops_info.push [loop, :hole, interiors[i+1]] }
	pshape.loops[1..-1].each_with_index { |loop, i| loops_info.push [loop.reverse, :hole, interiors[i+1]] }
	loops_info
end

#Check if a point is inside the stencil shape
def point_inside?(pt2d, on_border=true)
	@lst_pshapes.each do |pshape|	
		status = pseudoshape_point_inside?(pshape, pt2d, on_border)
		return true if status
	end
	false
end

def pseudoshape_point_inside?(pshape, pt2d, on_border=true)
	loops = pshape.loops
	if loops.length > 1
		if Geom.point_in_polygon_2D(Geom::Point3d.new(pt2d.x, pt2d.y, 0), pshape.main_loop, on_border)
			return true unless loops[1..-1].find { |loop| Geom.point_in_polygon_2D(Geom::Point3d.new(pt2d.x, pt2d.y, 0), loop, !on_border) }
		end	
	else
		return true if Geom.point_in_polygon_2D(Geom::Point3d.new(pt2d.x, pt2d.y, 0), pshape.main_loop, on_border)		
	end
	false
end

#INFO: Check if a point is inside the stencil shape
def face_inside?(face, t2d)
	pts2d = face.vertices.collect { |vx| p = t2d * vx.position ; p.z = 0 ; p }
	@lst_pshapes.each do |pshape|	
		return true if pseudoshape_face_inside?(pshape, face, t2d)
	end
	false
end

def pseudoshape_face_inside?(pshape, face, t2d)
	ls_pts2d = face.loops.collect { |loop| loop.vertices.collect { |vx| p = t2d * vx.position ; p.z = 0 ; p } }
	pseudoshape_contour_inside_main?(pshape, ls_pts2d)
end

def pseudoshape_contour_inside_main?(pshape, loop_pts2d)
	#Check if inside outer loop
	loop_pts2d.each do |pts2d|
		pts2d.each do |pt2d|
			return false unless pseudoshape_point_inside_main?(pshape, pt2d)
		end
		for i in 0..pts2d.length-2
			ptmid = Geom.linear_combination 0.5, pts2d[i], 0.5, pts2d[i+1]
			return false unless pseudoshape_point_inside_main?(pshape, ptmid)		
		end
	end
	
	#Check if not inside holes
	loop_pts2d.each do |pts2d|
		pts2d.each do |pt2d|
			return true unless pseudoshape_point_inside_hole?(pshape, pt2d)
		end
	end	
	false
end

def pseudoshape_point_inside_main?(pshape, pt2d)
	main_loop = pshape.main_loop
	return false unless Geom.point_in_polygon_2D(pt2d, main_loop, true)
	return true if pshape.loops.length == 1
	
	pshape.loops[1..-1].each do |loop|
		return false if Geom.point_in_polygon_2D(pt2d, loop, false)
	end
	true
end

def pseudoshape_point_inside_hole?(pshape, pt2d)
	pt2d = Geom::Point3d.new(pt2d.x, pt2d.y, 0)
	return false if pshape.loops.length == 1
	
	pshape.loops[1..-1].each do |loop|
		return true if Geom.point_in_polygon_2D(pt2d, loop, true)
	end
	false
end

#INFO: Check if a point is inside the stencil shape: true, false or nil
def pseudoshape_point_interior_status(pshape, pt2d)
	pshape.loops.each_with_index do |loop, iloop| 
		interiors = pshape.interiors[iloop]
		loop.each_with_index { |pt, ipt| return interiors[ipt] if pt == pt2d }
	end
	true
end

#INFO: Check if a point is inside the stencil shape: true, false or nil
def pseudoshape_line_interior_status(pshape, line2d)
	pshape.loops.each_with_index do |loop, iloop| 
		interiors = pshape.interiors[iloop]
		loop.each_with_index { |pt, ipt| return interiors[ipt] if pt.on_line?(line2d) }
	end
	:unknown
end

#---------------------------------------------------------------------------------------------
# BBOX: Bounding Box for Stencil
#---------------------------------------------------------------------------------------------

#BBOX: Compute the normalized bounding box
def bbox_compute
	@bbox = Geom::BoundingBox.new
	
	@lst_pshapes.each do |pshape|
		@bbox.add pseudoshape_get_anchor(pshape)
		pshape.main_loop.each { |pt| @bbox.add pt }
	end
	
	@tr_bbox = Geom::Transformation.translation @bbox.min.vector_to(ORIGIN)
end

#BBOX: return the dimensions of the bounding box
def bbox_dimensions(base=nil)
	bbox_compute unless @bbox
	w = @bbox.width
	h = @bbox.height
	w = 0.1 if w == 0
	h = 0.1 if h == 0
	[w, h]
end

#BBOX: OpenGL Instructions for drawing
def bbox_instructions(dx, dy, bkcolor=nil, frcolor=nil)
	w, h = bbox_dimensions
	
	#Calculating the scaling
	s = [dx / w, dy / h].min
	ts = Geom::Transformation.scaling s, s, 1
	t = ts * @tr_bbox
	
	#Computing the OpenGL instructions
	bkcolor = 'yellow' unless bkcolor
	frcolor = 'darkgray' unless frcolor
	instructions = []
	@lst_pshapes.each do |pshape|
		@bbox.add pseudoshape_get_anchor(pshape)
		pshape.polys.each do |poly|
			pts = poly.collect { |pt| t * pt }
			instructions.push [GL_POLYGON, pts, bkcolor] if pts.length > 2
		end	
		pshape.loops.each do |loop| 
			pts = loop.collect { |pt| t * pt }
			instructions.push [GL_LINE_LOOP, pts, frcolor, 1, ''] if pts.length > 2
		end
	end
	
	#Individual anchors
	@lst_pshapes.each do |pshape|
		pt = t * pseudoshape_get_anchor(pshape)
		pts = G6.pts_square pt.x, pt.y, 1
		instructions.push [GL_POLYGON, pts, 'blue']
	end
	
	#Main Anchor
	anchor = t * ORIGIN
	pts = G6.pts_circle anchor.x, anchor.y, 4
	instructions.push [GL_POLYGON, pts, 'red']
	
	instructions
end

#---------------------------------------------------------------------------------------------
# DRAW: Manage the drawing of the stencil
#---------------------------------------------------------------------------------------------

def draw_flat(view, tr, bkcolor, frcolor, width, stipple, small_offset=nil)
	@lst_pshapes.each do |pshape|
		pshape.loops.each do |loop|
			pts = loop.collect { |pt| tr * pt }
			pts = pts.collect { |pt| G6.small_offset view, pt, small_offset } if small_offset
			view.line_width = width
			view.line_stipple = stipple
			view.drawing_color = bkcolor
			view.draw GL_POLYGON, pts
			view.drawing_color = frcolor
			view.draw GL_LINE_LOOP, pts
		end
	end	
end

#Draw the stencil in 2.5D 
def draw_flat2d(view, tr, bkcolor, frcolor, width, stipple, small_offset=nil)
	view.line_width = width
	view.line_stipple = stipple
	view.drawing_color = bkcolor
	
	#Drawing the polygon for each shape
	@lst_pshapes.each do |pshape|
		pshape.polys.each do |poly|
			pts = poly.collect { |pt| tr * pt }
			pts2d = pts.collect { |pt| view.screen_coords(pt) }
			view.draw2d GL_POLYGON, pts2d		
		end
	end	
		
	#Drawing the contours for each shape	
	@lst_pshapes.each do |pshape|
		pshape.loops.each do |loop|
			pts = loop.collect { |pt| tr * pt }
			pts2d = pts.collect { |pt| view.screen_coords(pt) }
			view.drawing_color = frcolor
			view.draw2d GL_LINE_LOOP, pts2d if pts2d.length > 2
		end
	end	
	
	#Drawing a mark at centroid of each individual shape
	@lst_pshapes.each do |pshape|
		pt = tr * pseudoshape_get_anchor(pshape)
		pt2d = view.screen_coords(pt)
		pts2d = G6.pts_square pt2d.x, pt2d.y, 1
		view.drawing_color = 'blue'
		view.draw2d GL_POLYGON, pts2d
	end		
end

end	#class Stencil

end #module Traductor