=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed December 2014 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6Triangulation.rb
# Original Date	:   06 Dec 2014
# Description	:   Implement triangulation algorithms (body)
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#================================================================================		
#================================================================================
# Class DelaunayTriangulation2D	: Straight Delaunay Triangulation in 2D
#
# Credit to Paul Bourke (pbourke@swin.edu.au) for algorithm
#================================================================================		
#================================================================================	

class DelaunayTriangulation2D

#Class instance initialization
def initialize__(*hargs)
	hargs.each { |harg| harg.each { |key, value| parse_args(key, value) } if harg.class == Hash }
end

#Parsing the arguments 
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /progression_proc/i
		@progression_proc = value
	when /progression_number/i
		@progression_number = value
	end	
end
	
#Top calculation method
def calculate(lpt, continue=false)
	@tbeg = Time.now
	
	#Initialize the Delaunay diagram or continue with new points
	(@npt && continue) ? diagram_continue(lpt) : diagram_init(lpt)
	
	#Parameters for notification of progression
	if @progression_proc
		@ipt_beg = @range_points.first
		@npt_calc = @range_points.last - @ipt_beg + 1
		@npt_slice = @progression_number
		@npt_slice = 500 unless @npt_slice.class == Integer
	end
	
	#Adding the cloud points one by one
	@range_points.each { |ipt| insert_new_point ipt }
	
	#Returning the triangles with point indexes remapped, eliminating triangles with points of supertriangle
	lmap = @lst_ptxyf.collect { |a| a[3] }
	itriangles = []
	for i in 0..@ntriangles
		next if @lst_invalid[i]
		tri, = @lst_triangles_info[i]
		itriangles.push tri.collect { |j| lmap[j] } if tri && (tri & @isupers).empty?
	end	
	itriangles
end	

#Inserting a new point in the diagram
def insert_new_point(ipt)
	if @progression_proc && (ipt - @ipt_beg).modulo(@npt_slice) == 0 
		ratio = 1.0 * (ipt - @ipt_beg) / @npt_calc
		@progression_proc.call ratio
	end
	
	#Find the triangles which include the new point within their circumcircle
	xp, yp = @lst_ptxyf[ipt]
	hsh_borders = {}
	ntri = @ntriangles-1
	for itri in 0..ntri
		next if @lst_invalid[itri] || @lst_complete[itri]
		tri_info = @lst_triangles_info[itri]
		tri, tri_pt, center, radius2 = tri_info
		#next unless tri
		
		#Checking if point is within circumcircle (circumcircle is computed on the fly if not already)
		xc, yc = center
		xd = xp - xc 
		@lst_complete[itri] = true if xd > 0 && xd * xd > radius2
		next unless (xd * xd + (yc - yp) * (yc - yp)) <= radius2 #+ 0.00001
		
		#Removing the triangle <itri> from the list by declaring it invalid
		@lst_invalid[itri] = true
		
		#Adding the border of the triangle in the list. If duplicate, cancel it, so we keep outer borders only
		ipt1, ipt2, ipt3 = tri
		[[ipt1, ipt2], [ipt2, ipt3], [ipt3, ipt1]].each do |i, j|
			key = (i < j) ? i * 110000 + j : j * 110000 + i
			(hsh_borders.has_key?(key)) ? hsh_borders.delete(key) : hsh_borders[key] = [i, j, itri]
		end		
	end
	
	#Creating the new triangles from the perimeter edges and the new point
	hsh_borders.values.each do |ipt1, ipt2, it| 
		i = register_triangle ipt1, ipt2, ipt
		@lst_complete[i] = true if i && @lst_complete[it]
	end	
end

#Initializing the Triangulation diagram. Input as list of Geom Points or [x, y] in float
def diagram_init(lpt)
	#Initializing the global variables describing the diagram
	@lst_triangles_info = []
	@lst_invalid = []
	@lst_complete = []
	@ntriangles = 0
	@npt = lpt.length
	@range_points = 0..@npt-1
	
	#Registering the Cloud Points with conversion to Float (much faster!)and sorting by x
	i = -1
	@lst_ptxyf = lpt.collect { |pt| i += 1 ; [pt.x.to_f, pt.y.to_f, pt.z.to_f, i] }
	@lst_ptxyf = @lst_ptxyf.sort
	
	#Computing the Bounding box to all points
	xmin = @lst_ptxyf[0][0]
	xmax = @lst_ptxyf[-1][0]
	ymin = ymax = @lst_ptxyf[0][1]
	@lst_ptxyf.each do |x, y|
		ymin = y if y < ymin
		ymax = y if y > ymax
	end	
	fac = 6
	dx = (xmin == xmax) ? fac : fac * (xmax - xmin)
	dy = (ymin == ymax) ? fac : fac * (ymax - ymin)
		
	#Creating the initial Super Triangle
	@lst_ptxyf.push [xmin - dx, ymin - dy], [xmin + dx * 1.1, ymin - dy], [(xmax + xmin) / 2, ymax + dy]
	@isupers = [@npt, @npt+1, @npt+2]
	register_triangle *@isupers
end	

#Initializing the Triangulation diagram. Input as list of Geom Points or [x, y] in float
def diagram_continue(lpt)
	#Initializing the global variables describing the diagram
	ls = []
	@lst_triangles_info.each_with_index { |a, i| ls.push a unless @lst_invalid[i] }
	@lst_triangles_info = ls
	@ntriangles = @lst_triangles_info.length
	@lst_invalid = []
	@lst_complete = []
	
	#Registering the Cloud Points with conversion to Float (much faster!) and sorting by x
	npt = @lst_ptxyf.length
	@range_points = npt..npt+lpt.length-1
	i = npt - 4
	new_ptxyf = lpt.collect { |pt| i += 1 ; [pt.x.to_f, pt.y.to_f, pt.z.to_f, i] }
	@lst_ptxyf.concat new_ptxyf.sort
end

#Registering a triangle and computing its circumcircle
def register_triangle(i0, i1, i2)
	itri = @ntriangles
	tri_pt = [@lst_ptxyf[i0], @lst_ptxyf[i1], @lst_ptxyf[i2]]
	center, radius2 = compute_circumcircle(*tri_pt)
	return nil unless center
	@lst_triangles_info[itri] = [[i0, i1, i2], tri_pt, center, radius2]
	@ntriangles += 1
	itri
end

#Compute the circumcircle (center, square radius) to 3 points [Source Wikipedia]
def compute_circumcircle(pt1, pt2, pt3)
	x1, y1 = pt1
	x2, y2 = pt2
	x3, y3 = pt3
	
	yd12 = y1 - y2
	yd23 = y2 - y3
	yd31 = y3 - y1
	d = (x1 * yd23 + x2 * yd31 + x3 * yd12) * 2
	return nil if d == 0
	
	sqd1 = x1 * x1 + y1 * y1
	sqd2 = x2 * x2 + y2 * y2
	sqd3 = x3 * x3 + y3 * y3
	xc = (sqd1 * yd23 + sqd2 * yd31 + sqd3 * yd12) / d 
	yc = (sqd1 * (x3 - x2) + sqd2 * (x1 - x3) + sqd3 * (x2 - x1)) / d 
	
	dx = x2 - xc
	dy = y2 - yc
	[[xc, yc], dx * dx + dy * dy]
end

end	#class DelaunayTriangulation2D

#================================================================================		
#================================================================================
# Class ConcaveHull2D : Calculate a concave hull for a set of planar points
#================================================================================		
#================================================================================	

class ConcaveHull2D

#Class instance initialization
def initialize__(*hargs)
	@small_angle = 10
	@fac_distmax = 10
	@fact_distmin = 0.005
	hargs.each { |harg| harg.each { |key, value| parse_args(key, value) } if harg.class == Hash }
end

#Parsing the arguments 
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /progression_proc/i
		@progression_proc = value
	when /progression_number/i
		@progression_number = value
	when /small_angle/i
		@small_angle = value
	when /fac_distmax/i
		@fac_distmax = value
	when /fact_distmin/i
		@fact_distmin = value
	end	
end

def progression_delaunay(ratio)
	@progression_proc.call ratio if @progression_proc
end

#Top method to compute the Concave hull
def compute(lpt, *hargs)
	hargs.each { |harg| harg.each { |key, value| parse_args(key, value) } if harg.class == Hash }
	@limit_angle = 2 * @small_angle
	@limit_neg_angle = 2 * @limit_angle

	@lpt = lpt
	
	#Dimension of the cloud
	box = Geom::BoundingBox.new
	box.add lpt
	diag = box.diagonal
	@distmin = diag * @fact_distmin
	
	#Creating the Delaunay algorithm class instance
	hsh = { :progression_number => 1000, :progression_proc => self.method("progression_delaunay") }
	@delaunay_algo = Traductor::DelaunayTriangulation2D.new hsh

	#Calculating the triangulation
	@lst_itriangles = @delaunay_algo.calculate(lpt)
	
	#Registering the borders
	@hsh_tri_invalid = {}
	@lst_points_info = []	
	@hsh_borders_info = {}	
	@hsh_borders_ignored = {}	
	@hsh_borders_bad = {}	
	@lst_itriangles.each_with_index do |itri_pt, itri|
		ip1, ip2, ip3 = itri_pt
		register_borders(ip1, ip2, itri)
		register_borders(ip2, ip3, itri)
		register_borders(ip3, ip1, itri)
	end
	
	#Algorithm to build the concave hull by progressive elimination of borders
	iterate_concave
	
	#Computing the outer borders of the triangulation
	hsh_outer_borders = compute_outer_borders
	borders_to_contour(hsh_outer_borders)
end

#Algorithm to explore and remove borders
def iterate_concave
	n = 0
	while true
		n += 1
		break if n > 4000
		ls_bad_borders = []
		hsh_outer_borders = compute_outer_borders
		lsborders = []
		hsh_outer_borders.each { |key, border| lsborders.push border unless @hsh_borders_ignored[key] }
		lsborders = lsborders.sort { |a, b| b[4] <=> a[4] }
		lsborders.each do |border|
			itri, i, j, key = border
			@hsh_borders_ignored[key] = true
			next if true_number_border(i) < 3 || true_number_border(j) < 3
			if borders_bad?(border, hsh_outer_borders)
				ls_bad_borders.push border
				@hsh_tri_invalid[itri] = true
				@hsh_borders_bad[key] = true
				break
			end	
		end	
		break if ls_bad_borders.empty?
	end
end

#Compute the actual number of borders at a given step
def true_number_border(ipt)
	lkeys = @lst_points_info[ipt].uniq
	nsb = 0
	lkeys.each do |key|
		nsb += 1 if @hsh_borders_info[key].find { |b| !@hsh_tri_invalid[b[0]] }
	end	
	nsb
end

#Compute the outer borders
def compute_outer_borders
	hsh_outer_borders = {}
	@hsh_borders_info.each do |key, ls|
		ls = ls.find_all { |a| !@hsh_tri_invalid[a[0]] }
		hsh_outer_borders[key] = ls[0] if ls.length == 1
	end	
	hsh_outer_borders
end

#Register a border
def border_key(i, j) ; (i < j) ? i * 110000 + j : j * 110000 + i ; end
def register_borders(i, j, itri)
	key = border_key(i, j)
	ls = @hsh_borders_info[key]
	ls = @hsh_borders_info[key] = [] unless ls
	ls.push [itri, i, j, key, @lpt[i].distance(@lpt[j])]
	@lst_points_info[i] = [] unless @lst_points_info[i]
	@lst_points_info[i].push key
	@lst_points_info[j] = [] unless @lst_points_info[j]
	@lst_points_info[j].push key
end

#Check if borders is with a small angle
def borders_bad?(border, hsh_outer_borders)
	itri, i, j = border
	return false unless itri
	ip1, ip2, ip3 = @lst_itriangles[itri]
	ip = [ip1, ip2, ip3].find { |k| k != i && k != j }
	
	#Triangle has 2 outer borders - Ignore it
	key_i = border_key(i, ip)
	lpki = @lst_points_info[i].find_all { |key| hsh_outer_borders.has_key?(key) }
	return false if hsh_outer_borders.has_key?(key_i) && lpki.length <= 2
	key_j = border_key(j, ip)
	lpkj = @lst_points_info[j].find_all { |key| hsh_outer_borders.has_key?(key) }
	return false if hsh_outer_borders.has_key?(key_j) && lpkj.length <= 2
	
	pti = @lpt[i]
	ptj = @lpt[j]
	ptn = @lpt[ip]
	vecij = pti.vector_to ptj
	vecin = pti.vector_to ptn
	vecjn = ptj.vector_to ptn
	anglei = vecij.angle_between(vecin).radians
	anglej = vecij.reverse.angle_between(vecjn).radians
	dij = pti.distance(ptj)
	
	status = false
	
	if (anglei < @small_angle && anglej < @limit_angle) || 
	   (anglej < @small_angle && anglei < @limit_angle)
		return true
	elsif (anglei < @small_angle && anglej < @limit_neg_angle && ptj.distance(ptn) < @distmin) || 
	      (anglej < @small_angle && anglei < @limit_neg_angle && pti.distance(ptn) < @distmin)
		return true
	elsif dij > @distmin * 5
		dmin_i = min_distance_at_point(i)
		dmin_j = min_distance_at_point(j)
		#dmin = (dmin_i + dmin_j) * 0.5
		dmin = [[dmin_i, dmin_j].min, @distmin * 0.8].max
		return true if dmin * @fac_distmax < dij
	end	
	
	status
end

#Compute the minimum distance at a point to neighbours points
def min_distance_at_point(ipt)
	lborders = @lst_points_info[ipt].collect { |key| @hsh_borders_info[key] }
	dmin = nil
	lborders.each do |border|
		border.each do |itri, i, j|
			d = @lpt[i].distance @lpt[j]
			dmin = d if !dmin || d < dmin
		end	
	end	
	dmin
end

#Compute the contours from borders
def borders_to_contour(hsh_outer_borders)
	contours = []
	while hsh_outer_borders.length > 0
		border0 = hsh_outer_borders.shift
		key0, a = border0
		itri, ibeg, inext = a
		contour = [ibeg, inext]
		
		n = 0
		while true
			n += 1
			break if n > 10000
			key_border = @lst_points_info[inext].find { |key| hsh_outer_borders.has_key?(key) }
			border = hsh_outer_borders[key_border]
			hsh_outer_borders.delete key_border
			itri, i, j = border
			inext = (i == inext) ? j : i
			break if inext == ibeg
			contour.push inext
		end
		contours.push contour.collect { |i| @lpt[i] }
	end	
	contours
end

end	#ConcaveHull2D

end	#module Traductor
